"""
Module name
"""

from __future__ import print_function

import sys


class ProbLogError(Exception):
    """General Problog error. Root of all ProbLog errors that can be blamed on user input.

    :param message: error message
    :param location: location at which the error occurred (referring to an user input file)
    """

    def __init__(self, message, location=None, **extra):
        self.errtype = self.__class__.__name__
        self.base_message = message
        self.location = location
        self.message = self._message()
        for k, v in extra.items():
            setattr(self, k, v)

    def _location_string(self):
        if self.location is None:
            return ''
        if type(self.location) == tuple:
            if len(self.location) != 3:
                return ''
            fn, ln, cn = self.location
            if fn is None:
                return ' at %s:%s' % (ln, cn)
            else:
                return ' at %s:%s in %s' % (ln, cn, fn)
        else:
            return ' at character %s' % self.location

    def _message(self):
        return '%s%s.' % (self.base_message, self._location_string())

    def __str__(self):
        return self.message


class ParseError(ProbLogError):
    """Error during parsing."""
    pass


class GroundingError(ProbLogError):
    """Represents an error that occurred during grounding."""

    pass


class CompilationError(ProbLogError):
    """Error during compilation"""
    pass


class InstallError(ProbLogError):
    """Error during installation"""
    pass


class InvalidValue(ProbLogError):
    pass


class UserError(ProbLogError):
    pass


class InconsistentEvidenceError(ProbLogError):
    """Error when evidence is inconsistent"""

    def __init__(self, source=None, context=''):
        """

        :param source: evidence term that causes the problem
        :param context: extra message describing the context (e.g. example number in lfi)
        :return:
        """
        self.source = source
        self.context = context
        if source is None:
            ProbLogError.__init__(self, "Inconsistent evidence detected%s" % context)
        else:
            ProbLogError.__init__(self, "Inconsistent evidence detected%s: '%s'" % (context, source))


def process_error(err, debug=False):
    """Take the given error raise by ProbLog and produce a meaningful error message.

    :param err: error that was raised
    :param debug: if True, also print original stack trace
    :return: textual representation of the error
    """
    if debug and hasattr(err, 'trace'):
        print(err.trace, file=sys.stderr)

    if isinstance(err, ProbLogError):
        return '%s: %s' % (err.__class__.__name__, err)
    elif 'Timeout' in str(err) or 'timeout' in str(err):
        return 'Timeout exceeded'
    elif isinstance(err, KeyboardInterrupt):
        return 'Interrupted by user'
    else:
        if not debug and hasattr(err, 'trace'):
            print(err.trace, file=sys.stderr)
        return 'An unexpected error has occurred.'



"""
problog.sdd_formula - Sentential Decision Diagrams
--------------------------------------------------

Interface to Sentential Decision Diagrams (SDD)

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from __future__ import print_function

from .formula import LogicDAG, LogicFormula
from .core import transform
from .errors import InstallError
from .dd_formula import DD, build_dd, DDManager
from .util import mktempfile
import os


# noinspection PyBroadException
try:
    import sdd
except Exception as err:
    sdd = None


class SDD(DD):
    """A propositional logic formula consisting of and, or, not and atoms represented as an SDD.

    This class has two restrictions with respect to the default LogicFormula:

        * The number of atoms in the SDD should be known at construction time.
        * It does not support updatable nodes.

    This means that this class can not be used directly during grounding.
    It can be used as a target for the ``makeAcyclic`` method.
    """

    transform_preference = 10

    def __init__(self, sdd_auto_gc=False, **kwdargs):
        if sdd is None:
            raise InstallError('The SDD library is not available. Please run the installer.')
        self.auto_gc = sdd_auto_gc
        DD.__init__(self, auto_compact=False, **kwdargs)

    def _create_manager(self):
        return SDDManager(auto_gc=self.auto_gc)

    @classmethod
    def is_available(cls):
        """Checks whether the SDD library is available."""
        return sdd is not None

    def to_formula(self):
        """Extracts a LogicFormula from the SDD."""
        formula = LogicFormula()

        for n, q, l in self.labeled():
            node = self.get_inode(q)
            constraints = self.get_constraint_inode()
            nodec = self.get_manager().conjoin(node, constraints)
            i = self._to_formula(formula, nodec)
            formula.add_name(n, i, formula.LABEL_QUERY)
        return formula

    def _to_formula(self, formula, current_node):
        if self.get_manager().is_true(current_node):
            return formula.TRUE
        elif self.get_manager().is_false(current_node):
            return formula.FALSE
        elif sdd.sdd_node_is_literal(current_node):  # it's a literal
            lit = sdd.sdd_node_literal(current_node)
            at = self.var2atom[abs(lit)]
            node = self.get_node(at)
            if lit < 0:
                return -formula.add_atom(-lit, probability=node.probability, name=node.name, group=node.group)
            else:
                return formula.add_atom(lit, probability=node.probability, name=node.name, group=node.group)
        else:  # is decision
            size = sdd.sdd_node_size(current_node)
            elements = sdd.sdd_node_elements(current_node)
            primes = [sdd.sdd_array_element(elements, i) for i in range(0, size * 2, 2)]
            subs = [sdd.sdd_array_element(elements, i) for i in range(1, size * 2, 2)]

            # Formula: (p1^s1) v (p2^s2) v ...
            children = []
            for p, s in zip(primes, subs):
                p_n = self._to_formula(formula, p)
                s_n = self._to_formula(formula, s)
                c_n = formula.add_and((p_n, s_n))
                children.append(c_n)
            return formula.add_or(children)

    def to_dot(self, *args, **kwargs):
        return self.to_formula().to_dot(*args, **kwargs)


class SDDManager(DDManager):
    """
    Manager for SDDs.
    It wraps around the SDD library and offers some additional methods.
    """

    def __init__(self, varcount=0, auto_gc=True):
        """Create a new SDD manager.

        :param varcount: number of initial variables
        :type varcount: int
        :param auto_gc: use automatic garbage collection and minimization
        :type auto_gc: bool
        """
        DDManager.__init__(self)
        if varcount is None or varcount == 0:
            varcount = 1
        self.__manager = sdd.sdd_manager_create(varcount, auto_gc)
        self.varcount = varcount

    def get_manager(self):
        """Get the underlying sdd manager."""
        return self.__manager

    def add_variable(self, label=0):
        if label == 0 or label > self.varcount:
            sdd.sdd_manager_add_var_after_last(self.__manager)
            self.varcount += 1
            return self.varcount
        else:
            return label

    def literal(self, label):
        self.add_variable(abs(label))
        return sdd.sdd_manager_literal(label, self.__manager)

    def is_true(self, node):
        return sdd.sdd_node_is_true(node)

    def true(self):
        return sdd.sdd_manager_true(self.__manager)

    def is_false(self, node):
        return sdd.sdd_node_is_false(node)

    def false(self):
        return sdd.sdd_manager_false(self.__manager)

    def conjoin2(self, a, b):
        return sdd.sdd_conjoin(a, b, self.__manager)

    def disjoin2(self, a, b):
        return sdd.sdd_disjoin(a, b, self.__manager)

    def negate(self, node):
        new_sdd = sdd.sdd_negate(node, self.__manager)
        self.ref(new_sdd)
        return new_sdd

    def same(self, node1, node2):
        # Assumes SDD library always reuses equivalent nodes.
        if node1 is None or node2 is None:
            return node1 == node2
        else:
            return int(node1) == int(node2)

    def ref(self, *nodes):
        for node in nodes:
            sdd.sdd_ref(node, self.__manager)

    def deref(self, *nodes):
        for node in nodes:
            sdd.sdd_deref(node, self.__manager)

    def write_to_dot(self, node, filename):
        sdd.sdd_save_as_dot(filename, node)

    def wmc(self, node, weights, semiring):
        logspace = 0
        if semiring.one() == 0.0:
            logspace = 1
        wmc_manager = sdd.wmc_manager_new(node, logspace, self.get_manager())
        varcount = sdd.sdd_manager_var_count(self.get_manager())

        # varlist_intern = sdd.sdd_variables(node, self.get_manager())
        # vc = sdd.sdd_manager_var_count(self.get_manager()) + 1
        # varlist = [i for i in range(0, vc) if sdd.sdd_array_int_element(varlist_intern, i)]

        for n in weights:
            pos, neg = weights[n]
            if n <= varcount:
                sdd.wmc_set_literal_weight(n, pos, wmc_manager)  # Set positive literal weight
                sdd.wmc_set_literal_weight(-n, neg, wmc_manager)  # Set negative literal weight
        result = sdd.wmc_propagate(wmc_manager)
        sdd.wmc_manager_free(wmc_manager)
        return result

    def wmc_literal(self, node, weights, semiring, literal):
        logspace = 0
        if semiring.one() == 0.0:
            logspace = 1
        wmc_manager = sdd.wmc_manager_new(node, logspace, self.get_manager())
        varcount = sdd.sdd_manager_var_count(self.get_manager())
        for i, n in enumerate(sorted(weights)):
            i += 1
            pos, neg = weights[n]
            if n <= varcount:
                sdd.wmc_set_literal_weight(n, pos, wmc_manager)  # Set positive literal weight
                sdd.wmc_set_literal_weight(-n, neg, wmc_manager)  # Set negative literal weight
        sdd.wmc_propagate(wmc_manager)

        result = sdd.wmc_literal_pr(literal, wmc_manager)
        sdd.wmc_manager_free(wmc_manager)
        return result

    def wmc_true(self, weights, semiring):
        logspace = 0
        if semiring.one() == 0.0:
            logspace = 1
        wmc_manager = sdd.wmc_manager_new(self.true(), logspace, self.get_manager())
        varcount = sdd.sdd_manager_var_count(self.get_manager())
        for i, n in enumerate(sorted(weights)):
            i += 1
            pos, neg = weights[n]
            if n <= varcount:
                sdd.wmc_set_literal_weight(n, pos, wmc_manager)  # Set positive literal weight
                sdd.wmc_set_literal_weight(-n, neg, wmc_manager)  # Set negative literal weight
        result = sdd.wmc_propagate(wmc_manager)
        sdd.wmc_manager_free(wmc_manager)
        return result

    def __del__(self):
        # if sdd is not None and sdd.sdd_manager_free is not None:
        #     sdd.sdd_manager_free(self.__manager)
        self.__manager = None

    def __getstate__(self):
        tempfile = mktempfile()
        vtree = sdd.sdd_manager_vtree(self.get_manager())
        sdd.sdd_vtree_save(tempfile, vtree)
        with open(tempfile) as f:
            vtree_data = f.read()

        nodes = []
        for n in self.nodes:
            if n is not None:
                sdd.sdd_save(tempfile, n)

                with open(tempfile) as f:
                    nodes.append(f.read())
            else:
                nodes.append(None)

        sdd.sdd_save(tempfile, self.constraint_dd)
        with open(tempfile) as f:
            constraint_dd = f.read()

        os.remove(tempfile)
        return {'varcount': self.varcount, 'nodes': nodes, 'vtree': vtree_data, 'constraint_dd': constraint_dd}

    def __setstate__(self, state):
        self.nodes = []
        self.varcount = state['varcount']
        tempfile = mktempfile()
        with open(tempfile, 'w') as f:
            f.write(state['vtree'])
        vtree = sdd.sdd_vtree_read(tempfile)
        self.__manager = sdd.sdd_manager_new(vtree)

        for n in state['nodes']:
            if n is None:
                self.nodes.append(None)
            else:
                with open(tempfile, 'w') as f:
                    f.write(n)
                self.nodes.append(sdd.sdd_read(tempfile, self.__manager))

        with open(tempfile, 'w') as f:
            f.write(state['constraint_dd'])
        self.constraint_dd = sdd.sdd_read(tempfile, self.__manager)
        os.remove(tempfile)
        return


@transform(LogicDAG, SDD)
def build_sdd(source, destination, **kwdargs):
    """Build an SDD from another formula.

    :param source: source formula
    :param destination: destination formula
    :param kwdargs: extra arguments
    :return: destination
    """
    return build_dd(source, destination, **kwdargs)



"""
Conditional Probability Distributions (CPD) to expert ProbLog to a
Probabilistic Graphical Model (PGM).

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function

import itertools
from datetime import datetime
import re
from functools import reduce
import sys
import math
from collections import Counter, defaultdict, OrderedDict


class PGM(object):
    def __init__(self, directed=True, variables=None, factors=None):
        """Probabilistic Graphical Model."""
        self.directed = directed
        self.factors = OrderedDict()
        self.vars = OrderedDict()
        if variables:
            for var in variables:
                self.add_var(var)
        if factors:
            for factor in factors:
                self.add_factor(factor)

    def copy(self, **kwargs):
        return PGM(
            directed=kwargs.get('directed', self.directed),
            factors=kwargs.get('factors', self.factors.values()),
            variables=kwargs.get('variables', self.vars.values())
        )

    def add_var(self, var):
        self.vars[var.name] = var

    def add_factor(self, factor):
        """Add a CPD.
        :param factor: New factor to add to PGM
        """
        if self.directed:
            name = factor.rv
        else:
            name = factor.name
        assert (name is not None)
        factor.pgm = self
        if name in self.factors:
            self.factors[name] += factor
        else:
            self.factors[name] = factor

    def factors_topological(self):
        """Return the factors in a topological order."""
        # Links from parent-node to child-node
        if not self.directed:
            return self.factors.values()
        links = dict()
        for factor in self.factors.values():
            for parent in factor.parents:
                if parent in links:
                    links[parent].add(factor.rv)
                else:
                    links[parent] = {factor.rv}
        all_links = set(links.keys())
        nonroot = reduce(set.union, links.values())
        root = all_links - nonroot
        queue = root
        visited = set()
        factors = []
        while len(queue) > 0:
            for rv in queue:
                if rv in visited:
                    continue
                all_parents_visited = True
                try:
                    factor = self.factors[rv]
                except KeyError as exc:
                    print('Error: random variable has no CPD associated:')
                    print(exc)
                    sys.exit(1)
                for parent_rv in factor.parents:
                    if parent_rv not in visited:
                        all_parents_visited = False
                        break
                if all_parents_visited:
                    factors.append(factor)
                    visited.add(rv)
            queue2 = set()
            for rv in queue:
                if rv in links:
                    for next_rv in links[rv]:
                        queue2.add(next_rv)
            queue = queue2
        return factors

    def compress_tables(self, allow_disjunct=False):
        """Analyze CPTs and join rows that have the same probability
        distribution."""
        factors = []
        for idx, factor in enumerate(self.factors.values()):
            factors.append(factor.compress(allow_disjunct=allow_disjunct))
        return self.copy(factors=factors)

    def to_hugin_net(self):
        """Export PGM to the Hugin net format.
        http://www.hugin.com/technology/documentation/api-manuals
        """
        assert self.directed
        cpds = [cpd.to_factor() for cpd in self.factors_topological()]
        lines = ["%% Hugin Net format",
                 "%% Created on {}\n".format(datetime.now()),
                 "%% Network\n",
                 "net {",
                 "  node_size = (50,50);",
                 "}\n",
                 "%% Nodes\n"]
        lines += [cpd.to_huginnet_node() for cpd in cpds]
        lines += ["%% Potentials\n"]
        lines += [cpd.to_huginnet_potential() for cpd in cpds]
        return '\n'.join(lines)

    def to_xdsl(self):
        """Export PGM to the XDSL format defined by SMILE.
        https://dslpitt.org/genie/wiki/Appendices:_XDSL_File_Format_-_XML_Schema_Definitions
        """
        assert self.directed
        cpds = [cpd.to_factor() for cpd in self.factors_topological()]
        lines = ['<?xml version="1.0" encoding="ISO-8859-1" ?>',
                 '<smile version="1.0" id="Aa" numsamples="1000">',
                 '  <nodes>']
        lines += [cpd.to_xdsl_cpt() for cpd in cpds]
        lines += ['  </nodes>',
                  '  <extensions>',
                  '    <genie version="1.0" app="ProbLog" name="Network1" faultnameformat="nodestate">']
        lines += [cpd.to_xdsl_node() for cpd in cpds]
        lines += ['    </genie>',
                  '  </extensions>',
                  '</smile>']
        return '\n'.join(lines)

    def to_uai08(self):
        """Export PGM to the format used in the UAI 2008 competition.
        http://graphmod.ics.uci.edu/uai08/FileFormat
        """
        assert self.directed
        # do not sort topological such that reading in and writing out UAI format is identical (thus ordered by
        # variable name instead of topological)
        cpds = [cpd.to_factor() for cpd in self.factors.values()]
        number_variables = str(len(cpds))
        domain_sizes = [str(len(self.vars[cpd.rv].values)) for cpd in cpds]
        number_functions = str(len(cpds))
        lines = ['BAYES',
                 number_variables,
                 ' '.join(domain_sizes),
                 number_functions]
        lines += [cpd.to_uai08_preamble(cpds) for cpd in cpds]
        lines += ['']
        lines += [cpd.to_uai08_function() for cpd in cpds]
        return '\n'.join(lines)

    def to_problog(self, drop_zero=False, use_neglit=False, value_as_term=True, ad_is_function=False):
        """Export PGM to ProbLog.
        :param ad_is_function: Experimental
        :param value_as_term: Include the variable's value as the last term instead of as part of the predicate name
        :param use_neglit: Use negative literals if it simplifies the program
        :param drop_zero: Do not include head literals with probability zero
        """
        factors = [factor.to_factor() for factor in self.factors_topological()]
        lines = ["%% ProbLog program",
                 "%% Created on {}\n".format(datetime.now())]
        if self.directed:
            lines += [factor.to_problog(self, drop_zero=drop_zero,
                                        use_neglit=use_neglit,
                                        value_as_term=value_as_term,
                                        ad_is_function=ad_is_function) for factor in factors]
        else:
            lines += [factor.to_problog_undirected(self, drop_zero=drop_zero,
                                                   use_neglit=use_neglit,
                                                   value_as_term=value_as_term,
                                                   ad_is_function=ad_is_function) for factor in factors]
            # if ad_is_function:
            # lines += ["evidence(false_constraints,false)."]
        return '\n'.join(lines)

    def to_graphviz(self):
        """Export PGM to Graphviz dot format.
        http://www.graphviz.org
        """
        assert self.directed
        lines = ['digraph bayesnet {']
        for cpd in self.factors.values():
            lines.append('  {} [label="{}"];'.format(cpd.rv_clean(), cpd.rv))
            for p in cpd.parents:
                lines.append('  {} -> {}'.format(cpd.rv_clean(p), cpd.rv_clean()))
        lines += ['}']
        return '\n'.join(lines)

    def __str__(self):
        return '\n'.join([str(factor) for factor in self.factors.values()])


re_tounderscore = re.compile(r'[\(\),\]\[ ]')
re_toremove = re.compile(r"""[^a-zA-Z0-9_]""")

boolean_values = [
    ['0', '1'],
    [0, 1],
    ['f', 't'],
    ['false', 'true'],
    ['no', 'yes'],
    ['n', 'y'],
    ['neg', 'pos'],
    ['nay', 'aye'],
    [False, True]
]


class Variable(object):
    def __init__(self, name, values, detect_boolean=True, force_boolean=False, boolean_true=None):
        """Conditional Probability Distribution."""
        self.name = name
        self.values = values
        self.latent = False
        self.booleantrue = boolean_true  # Value that represents true
        if (force_boolean or detect_boolean) and len(self.values) == 2:
            for values in boolean_values:
                if (values[0] == self.values[0] and values[1] == self.values[1]) or \
                   (type(self.values[0]) == str and type(self.values[1]) == str and
                   values[0] == self.values[0].lower() and values[1] == self.values[1].lower()):
                    self.booleantrue = 1
                    break
                elif (values[1] == self.values[0] and values[0] == self.values[1]) or \
                     (type(self.values[0]) == str and type(self.values[1]) == str and
                     values[1] == self.values[0].lower() and values[0] == self.values[1].lower()):
                    self.booleantrue = 0
                    break
            if force_boolean and self.booleantrue is None:
                self.booleantrue = 1

    def clean(self, value=None):
        if value is None:
            rv = self.name
        else:
            rv = value
        rv = re_tounderscore.sub('_', rv)
        rv = re_toremove.sub('', rv)
        if not rv[0].islower():
            rv = rv[0].lower() + rv[1:]
            if not rv[0].islower():
                rv = "v" + rv
        return rv

    def to_problog_value(self, value, value_as_term=True):
        if self.booleantrue is None:
            if type(value) is frozenset and len(value) == 1:
                value, = value
            if type(value) is frozenset:
                # It is a disjunction of possible values
                if len(value) == len(self.values) - 1:
                    # This is the negation of one value
                    new_value, = frozenset(self.values) - value
                    if value_as_term:
                        return '\+' + self.clean() + '("' + str(new_value) + '")'
                    else:
                        return '\+' + self.clean() + '_' + self.clean(str(new_value))
                else:
                    if value_as_term:
                        return '(' + '; '.join(
                            [self.clean() + '("' + str(new_value) + '")' for new_value in value]) + ')'
                    else:
                        return '(' + \
                               '; '.join([self.clean() + '_' + self.clean(str(new_value)) for new_value in value]) + ')'
            else:
                if value_as_term:
                    return self.clean() + '("' + str(value) + '")'
                else:
                    return self.clean() + '_' + self.clean(str(value))
        else:
            # It is a Boolean atom
            if self.values[self.booleantrue] == value:
                return self.clean()
            elif self.values[1 - self.booleantrue] == value:
                return '\+' + self.clean()
            else:
                raise Exception('Unknown value: {} = {}'.format(self.name, value))

    def __str__(self):
        return '{}'.format(self.name)

    def copy(self, **kwargs):
        return Variable(
            name=kwargs.get('name', self.name),
            values=kwargs.get('values', self.values),
            boolean_true=kwargs.get('boolean_true', self.booleantrue)
        )


class Factor(object):
    def __init__(self, pgm, rv, parents, table, name=None, *args, **kwargs):
        """Conditional Probability Table with discrete probabilities.

        :param rv: random variable
        :param values: random variable domain
        :param parents: parent random variables
        :param table:

        a = Factor(pgm, 'a', [], [0.4,0.6])
        b = Factor(pgm, 'b', [a],
                   {('f',): [0.2,0.8],
                    ('t',): [0.7,0.3]})
        """
        self.rv = rv
        self.pgm = pgm
        self.name = name
        self.parents = parents
        if isinstance(table, list) or isinstance(table, tuple):
            self.table = {(): table}
        elif isinstance(table, dict):
            self.table = table
        else:
            raise ValueError('Unknown type (expected list, tuple or dict): {}'.format(type(table)))

    def to_factor(self):
        return self

    def copy(self, **kwargs):
        return Factor(
            rv=kwargs.get('rv', self.rv),
            pgm=kwargs.get('pgm', self.pgm),
            name=kwargs.get('name', self.name),
            parents=kwargs.get('parents', self.parents),
            table=kwargs.get('table', self.table)
        )

    def compress(self, allow_disjunct=False):
        """Table to tree using the ID3 decision tree algorithm.

        From:
            {('f', 'f'): (0.2, 0.8),
             ('f', 't'): (0.2, 0.8),
             ('t', 'f'): (0.1, 0.9),
             ('t', 't'): (0.6, 0.4)}
        To:
            {('f', None): (0.2, 0.8),
             ('t', 'f'):  (0.1, 0.9),
             ('t', 't'):  (0.6, 0.4)}
        """
        # Traverse through the tree
        # First tuple is path with no value assignment and all rows in the table
        table = []
        for k, v in self.table.items():
            # Tuples allow hashing for IG
            table.append((k, tuple(v)))
        nodes = [(tuple([None] * len(self.parents)), table)]
        new_table = {}
        cnt = 0
        while len(nodes) > 0:
            cnt += 1
            curpath, node = nodes.pop()
            if len(node) == 0:
                continue
            # All the same or all different? Then stop.
            k, v = zip(*node)
            c = Counter(v)
            if len(c.keys()) == len(v):
                for new_path, new_probs in node:
                    new_table[new_path] = new_probs
                continue
            if len(c.keys()) == 1:
                new_table[curpath] = node[0][1]
                continue
            # Find max information gain
            # ig_idx = self.maxinformationgainparent(node)
            ps = [cnt / len(v) for cnt in c.values()]  # Pr(x_i)
            h_cur = -sum([p * math.log(p, 2) for p in ps])  # Entropy
            igr_idx = None
            igr_max = -9999999
            for parent_idx, parent in enumerate(self.parents):
                # print('parent_idx:{}, parent:{}'.format(parent_idx, parent))
                if curpath[parent_idx] is not None:
                    continue
                bins = {}
                for value in self.pgm.vars[parent].values:
                    bins[value] = []
                for k, v in node:
                    bins[k[parent_idx]].append(v)
                ig = h_cur  # Information Gain
                iv = 0  # Intrinsic Value
                for bin_value, bin_labels in bins.items():
                    label_cnt = Counter(bin_labels)
                    ps = [cnt / len(bin_labels) for cnt in label_cnt.values()]  # Pr(x_i)
                    h = -sum([p * math.log(p, 2) for p in ps])  # Entropy
                    r = len(bin_labels) / len(node)
                    ig -= r * h
                    if r == 0:
                        iv -= 9999999
                    else:
                        iv -= r * math.log(r, 2)
                igr = ig / iv  # Information Gain Ratio
                # print('ig={:.4f}, iv={:.4f}, igr={:.4f}, hc={:.4f}, idx={}, parent={}'.format(ig, iv, igr, h_cur, parent_idx, self.parents[parent_idx]))
                if igr > igr_max:
                    igr_max = igr
                    igr_idx = parent_idx
            # Create next nodes
            if igr_idx is None:
                # No useful split found
                for new_path, new_probs in node:
                    new_table[new_path] = new_probs
                continue
            if allow_disjunct \
                    and len(self.pgm.vars[self.parents[igr_idx]].values) > max(2, len(c)):
                # If disjuncts are allowed, find values that lead to identical subtrees
                # and merge them
                branches = defaultdict(set)
                for value in self.pgm.vars[self.parents[igr_idx]].values:
                    probs = set()
                    values = set()
                    for parent_values, prob in node:
                        if parent_values[igr_idx] == value:
                            new_parent_values = [v for v in parent_values]
                            new_parent_values[igr_idx] = None
                            probs.add(tuple([tuple(new_parent_values), prob]))
                            values.add(value)
                    probs = frozenset(probs)
                    branches[probs].update(values)
                for probs, values in branches.items():
                    newpath = [v for v in curpath]
                    newpath[igr_idx] = frozenset(values)
                    newnode = [tuple(newpath), []]
                    for parent_values, prob in probs:
                        parent_values = list(parent_values)
                        parent_values[igr_idx] = frozenset(values)
                        newnode[1].append((tuple(parent_values), prob))
                    nodes.append(newnode)
            else:
                for value in self.pgm.vars[self.parents[igr_idx]].values:
                    newpath = [v for v in curpath]
                    newpath[igr_idx] = value
                    newnode = [tuple(newpath), []]
                    for parent_values, prob in node:
                        if parent_values[igr_idx] == value:
                            newnode[1].append((parent_values, prob))
                    nodes.append(newnode)
        return self.copy(table=new_table)

    def to_huginnet_node(self):
        rv = self.pgm.vars[self.rv]
        lines = ["node {} {{".format(rv.clean()),
                 "  label = \"{}\";".format(self.rv),
                 "  position = (100,100);",
                 "  states = ({});".format(' '.join(['"{}"'.format(v) for v in rv.values])),
                 "}\n"]
        return '\n'.join(lines)

    def to_huginnet_potential(self):
        rv = self.pgm.vars[self.rv]
        name = rv.clean()
        if len(self.parents) > 0:
            name += ' | ' + ' '.join([rv.clean(p) for p in self.parents])
        lines = ['potential ({}) {{'.format(name),
                 '  % ' + ' '.join([str(v) for v in rv.values]),
                 '  data = (']
        table = sorted(self.table.items())
        for k, v in table:
            lines.append('    ' + ' '.join([str(vv) for vv in v]) + ' % ' + ' '.join([str(kk) for kk in k]))
        lines += ['  );',
                  '}\n']
        return '\n'.join(lines)

    def to_xdsl_cpt(self):
        rv = self.pgm.vars[self.rv]
        lines = ['    <cpt id="{}">'.format(rv.clean())]
        for v in rv.values:
            lines.append('      <state id="{}" />'.format(v))
        if len(self.parents) > 0:
            lines.append('      <parents>{}</parents>'.format(' '.join([rv.clean(p) for p in self.parents])))
        table = sorted(self.table.items())
        probs = ' '.join([str(value) for k, values in table for value in values])
        lines.append('      <probabilities>{}</probabilities>'.format(probs))
        lines.append('    </cpt>')
        return '\n'.join(lines)

    def to_xdsl_node(self):
        rv = self.pgm.vars[self.rv]
        lines = [
            '      <node id="{}">'.format(rv.clean()),
            '        <name>{}</name>'.format(self.rv),
            '        <interior color="e5f6f7" />',
            '        <outline color="000080" />',
            '        <font color="000000" name="Arial" size="8" />',
            '        <position>100 100 150 150</position>',
            '      </node>']
        return '\n'.join(lines)

    def to_uai08_preamble(self, cpds):
        function_size = 1 + len(self.parents)
        rv_to_idx = {}
        for idx, cpd in enumerate(cpds):
            rv_to_idx[cpd.rv] = idx
        variables = [str(rv_to_idx[rv]) for rv in self.parents] + [str(rv_to_idx[self.rv])]
        return '{} {}'.format(function_size, ' '.join(variables))

    def to_uai08_function(self):
        rv = self.pgm.vars[self.rv]
        number_entries = str(len(self.table) * len(rv.values))
        lines = [number_entries]
        try:
            table = sorted(self.table.items())
        except TypeError:
            print('ERROR: Cannot convert compressed CPTs to UAI08 format')
            sys.exit(1)
        for k, v in table:
            lines.append(' ' + ' '.join([str(vv) for vv in v]))
        lines.append('')
        return '\n'.join(lines)

    def to_problog(self, pgm, drop_zero=False, use_neglit=False, value_as_term=True, ad_is_function=False):
        lines = []
        var = pgm.vars[self.rv]
        name = var.clean()
        # if len(self.parents) > 0:
        #   name += ' | '+' '.join([self.rv_clean(p) for p in self.parents])
        # table = sorted(self.table.items())
        table = self.table.items()
        # value_assignments = itertools.product(*[pgm.cpds[parent].value for parent in self.parents)

        line_cnt = 0
        for k, v in table:
            if var.booleantrue is not None and drop_zero and v[var.booleantrue] == 0.0 and not use_neglit:
                continue
            head_problits = []
            if var.booleantrue is None:
                for idx, vv in enumerate(v):
                    if not (drop_zero and vv == 0.0):
                        head_problits.append((vv, var.to_problog_value(var.values[idx], value_as_term)))
            else:
                if drop_zero and v[var.booleantrue] == 0.0 and use_neglit:
                    head_problits.append((None, var.to_problog_value(var.values[1 - var.booleantrue], value_as_term)))
                elif v[var.booleantrue] == 1.0:
                    head_problits.append((None, var.to_problog_value(var.values[var.booleantrue], value_as_term)))
                else:
                    head_problits.append((v[var.booleantrue],
                                          var.to_problog_value(var.values[var.booleantrue], value_as_term)))
            body_lits = []
            for parent, parent_value in zip(self.parents, k):
                if parent_value is not None:
                    parent_var = pgm.vars[parent]
                    body_lits.append(parent_var.to_problog_value(parent_value, value_as_term))

            if ad_is_function:
                for head_cnt, (head_prob, head_lit) in enumerate(head_problits):
                    if len(body_lits) == 0:
                        lines.append('({};1.0)::{}.'.format(head_prob, head_lit))
                    else:
                        # new_probfact = 'pf_{}_{}_{}'.format(self.rv_clean(), line_cnt, head_cnt)
                        # new_body_lits = body_lits + [new_probfact]
                        # lines.append('({};1.0)::{}.'.format(head_prob, new_probfact))
                        # lines.append('{} :- {}.'.format(head_lit, ', '.join(new_body_lits)))
                        lines.append('({};1.0)::{} :- {}.'.format(head_prob, head_lit, ', '.join(body_lits)))
            else:
                if len(body_lits) > 0:
                    body_str = ' :- ' + ', '.join(body_lits)
                else:
                    body_str = ''
                head_strs = []
                for prob, lit in head_problits:
                    if prob is None:
                        head_strs.append(str(lit))
                    else:
                        head_strs.append('{}::{}'.format(prob, lit))
                head_str = '; '.join(head_strs)
                lines.append('{}{}.'.format(head_str, body_str))
            line_cnt += 1

        if ad_is_function and var.booleantrue is None:
            head_lits = [var.to_problog_value(value, value_as_term) for value in var.values]
            # lines.append('false_constraints :- '+', '.join(['\+'+l for l in head_lits])+'.')
            lines.append('constraint([' + ', '.join(['\+' + l for l in head_lits]) + '], false).')
            for lit1, lit2 in itertools.combinations(head_lits, 2):
                # lines.append('false_constraints :- {}, {}.'.format(lit1, lit2))
                lines.append('constraint([{}, {}], false).'.format(lit1, lit2))

        return '\n'.join(lines)

    def to_problog_undirected(self, pgm, drop_zero=False, use_neglit=False, value_as_term=True, ad_is_function=False):
        lines = []
        name = self.name
        table = self.table.items()
        line_cnt = 0
        for k, v in table:
            assert (len(v) == 1)
            body_lits = []
            for parent, parent_value in zip(self.parents, k):
                if parent_value is not None:
                    parent_var = pgm.vars[parent]
                    body_lits.append(parent_var.to_problog_value(parent_value, value_as_term))

            body_str = ' :- ' + ', '.join(body_lits)
            head_str = '{}::{}({})'.format(v[0], name, line_cnt)
            lines.append('{}{}.'.format(head_str, body_str))
            line_cnt += 1

        return '\n'.join(lines)

    def __str__(self):
        lines = []
        try:
            table = sorted(self.table.items())
        except TypeError:
            table = self.table.items()
        for k, v in table:
            lines.append('{}: {}'.format(k, v))
        table = '\n'.join(lines)
        parents = ''
        if len(self.parents) > 0:
            parents = ', '.join(map(str, self.parents))

        if self.rv is not None:
            var = self.pgm.vars[self.rv]
            rv = var.name
            if len(self.parents) > 0:
                rv += ' | '
            return 'Factor ({}{}) = {}\n{}\n'.format(rv, parents, ', '.join([str(v) for v in var.values]), table)
        return 'Factor ({})\n{}\n'.format(parents, table)


class OrCPT(Factor):
    def __init__(self, pgm, rv, parentvalues=None):
        super(OrCPT, self).__init__(pgm, rv, set(), [])
        if parentvalues is None:
            self.parentvalues = []
        else:
            self.parentvalues = parentvalues
        self.parents.update([pv[0] for pv in self.parentvalues])

    def add(self, parentvalues):
        """Add list of tuples [('a', 1)].
        :param parentvalues: List of tuples (parent, index)
        """
        self.parentvalues += parentvalues
        self.parents.update([pv[0] for pv in parentvalues])

    def to_factor(self):
        rv = self.pgm.vars[self.rv]
        parents = sorted(list(set([pv[0] for pv in self.parentvalues])))
        table = dict()
        parent_values = [self.pgm.vars[parent].values for parent in parents]
        for keys in itertools.product(*parent_values):
            is_true = False
            for parent, value in zip(parents, keys):
                if (parent, value) in self.parentvalues:
                    is_true = True
            if is_true:
                table[keys] = [0.0, 1.0]
            else:
                table[keys] = [1.0, 0.0]
        return Factor(self.pgm, self.rv, rv.values, parents, table)

    def __add__(self, other):
        return OrCPT(self.pgm, self.rv, self.parentvalues + other.parentvalues)

    def __str__(self):
        rv = self.pgm.vars[self.rv]
        table = '\n'.join(['{}'.format(pv) for pv in self.parentvalues])
        parents = ''
        if len(self.parents) > 0:
            parents = ' -- {}'.format(','.join(self.parents))
        return 'OrCPT {} [{}]{}\n{}\n'.format(self.rv, ','.join(map(str, rv.values)), parents, table)



"""
Part of the ProbLog distribution.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import print_function

import unittest

from problog.program import PrologString
from problog.engine import DefaultEngine
from problog.logic import Term, Constant
from problog.formula import LogicFormula

import glob, os, random

class TestEngine(unittest.TestCase):

    def setUp(self) :

        try :
            self.assertCollectionEqual = self.assertItemsEqual
        except AttributeError :
            self.assertCollectionEqual = self.assertCountEqual


    def test_nonground_query_ad(self) :
        """Non-ground call to annotated disjunction"""

        program = """
            0.1::p(a); 0.2::p(b).
            query(p(_)).
        """

        engine = DefaultEngine()
        db = engine.prepare( PrologString(program) )

        result=None
        for query in engine.query(db, Term('query', None)) :
            result = engine.ground(db, query[0], result, label='query')

        found = [ str(x) for x, y in result.queries() ]

        self.assertCollectionEqual( found, [ 'p(a)', 'p(b)' ])


    def test_compare(self) :
        """Comparison operator"""

        program = """
            morning(Hour) :- Hour >= 6, Hour =< 10.
        """

        engine = DefaultEngine()
        db = engine.prepare( PrologString(program) )

        self.assertEqual( list(map(list,engine.query(db, Term('morning', Constant(8)) ))), [[8]])

    def test_anonymous_variable(self) :
        """Anonymous variables are distinct"""

        program = """
            p(_,X,_) :- X = 3.

            q(1,2,3).
            q(1,2,4).
            q(2,3,5).
            r(Y) :- q(_,Y,_).

        """

        engine = DefaultEngine()
        db = engine.prepare( PrologString(program) )
        self.assertEqual( list(map(list,engine.query(db, Term('p', Constant(1), Constant(3), Constant(2) )))), [[Constant(1),Constant(3),Constant(2)]])

        self.assertEqual(list(map(list,engine.query(db, Term('r', None )))), [[2], [3]])

    def test_functors(self) :
        """Calls with functors"""

        program = """
            p(_,f(A,B),C) :- A=y, B=g(C).
            a(X,Y,Z) :- p(X,f(Y,Z),c).
        """
        pl = PrologString(program)

        r1 = DefaultEngine().query(pl, Term('a',Term('x'),None,Term('g',Term('c'))))
        r1 = [ list(map(str,sol)) for sol in r1  ]
        self.assertCollectionEqual( r1, [['x', 'y', 'g(c)']])

        r2 = DefaultEngine().query(pl, Term('a',Term('x'),None,Term('h',Term('c'))))
        self.assertCollectionEqual( r2, [])

        r3 = DefaultEngine().query(pl, Term('a',Term('x'),None,Term('g',Term('z'))))
        self.assertCollectionEqual( r3, [])


class TestEngineCycles(unittest.TestCase):

    def setUp(self) :
        self.edges = [(1,3),(1,5),(1,6),(1,8),(2,3),(2,5),(2,8)]
        self.edges += [(4,5),(4,7),(4,9),(4,11),(5,9),(6,10),(6,11),(6,13)]
        self.edges += [(7,11),(7,12),(8,9),(8,10),(8,11),(9,11),(10,11),(11,14)]
        self.edges += [ (y,x) for x,y in self.edges ]
        self.edges = list(sorted(self.edges))
        self.nodes = set([ x for x,y in self.edges ] + [ y for x,y in self.edges ])

        self.clause_v1 = [ 'active(X) :- edge(X,Y), active(Y).' ]
        self.clause_v2 = [ 'active(X) :- active(Y), edge(X,Y).' ]
        self.clauses = [ 'active(X) :- active_p(X).', 'query(active(11)).', 'path_from(X) :- path(X,_).' ]

        self.facts = [ '0.2::active_p(%s).' % e for e in self.nodes ] + [ '0.3::edge(%s,%s).' % e for e in self.edges ]

        self.program_v1 = self.facts + self.clauses + self.clause_v1
        self.program_v2 = self.facts + self.clauses + self.clause_v2

        self.maxDiff = None
        try :
            self.assertCollectionEqual = self.assertItemsEqual
        except AttributeError :
            self.assertCollectionEqual = self.assertCountEqual



    def test_cycle_goodcode(self) :
        N = 20
        program = self.program_v1[:]

        for i in range(0,N) :
            seed = str(random.random())[2:]
            random.seed(seed)
            random.shuffle(program)
            txt = '\n'.join(program)
            f = DefaultEngine(label_all=True).ground_all(PrologString(txt))
            paths = list(list_paths(f))

            edges = set()
            for p in paths :
                for i in range(0, len(p)-1) :
                    edges.add( ( int(p[i]), int(p[i+1]) ) )
            edges = list(sorted(edges))

            # if (edges != self.edges) :
            #     with open('cycle_error.pl', 'w') as f :
            #         print(txt, file=f)
            #     with open('cycle_error.dot', 'w') as f :
            #         print('digraph CycleError {', file=f)
            #         for edge in edges :
            #             print('%s -> %s;' % edge, file=f)
            #         print('}', file=f)

            self.assertCollectionEqual(self.edges, edges, msg='Test failed for random seed %s' % seed)

    def test_cycle_badcode(self) :
        N = 20
        program = self.program_v2[:]

        for i in range(0,N) :
            seed = str(random.random())[2:]
            random.seed(seed)
            random.shuffle(program)
            txt = '\n'.join(program)
            # try :
            f = DefaultEngine(label_all=True).ground_all(PrologString(txt))
            # except Exception :
            #     with open('cycle_error.pl', 'w') as f :
            #         print(txt, file=f)
            #     raise


            paths = list(list_paths(f))

            edges = set()
            for p in paths :
                for i in range(0, len(p)-1) :
                    edges.add( ( int(p[i]), int(p[i+1]) ) )
            edges = list(sorted(edges))

            # if (edges != self.edges) :
            #     with open('cycle_error.pl', 'w') as f :
            #         print(txt, file=f)
            #     with open('cycle_error.dot', 'w') as f :
            #         print('digraph CycleError {', file=f)
            #         for edge in edges :
            #             print('%s -> %s;' % edge, file=f)
            #         print('}', file=f)


            self.assertCollectionEqual(self.edges, edges, msg='Test failed for random seed %s' % seed)


def list_paths(source) :
    names = dict( (index, str(source.get_name(index))) for index, node, tp in source if index != 0 )

    for name, node in source.queries() :
        for path in _list_paths(source, node, []) :
            path = [ names.get(n) for n in path ]
            path = [ int(p[7:-1]) for p in path if p != None and p.startswith('active(')  ]
            yield path
        break

from itertools import product
def _list_paths( source, node_id, path ) :
    if node_id in path :
        yield [node_id]
    else :
        add = [node_id]
        node = source.get_node(node_id)
        nodetype = type(node).__name__
        if nodetype == 'conj' :
            c1, c2 = node.children
            paths1 = list( _list_paths(source,c1, path + add ) )
            paths2 = list( _list_paths(source,c2, path + add ) )
            for p1,p2 in product( paths1, paths2 ) :
                yield add + p1 + p2
        elif nodetype == 'disj' :
            for c in node.children :
                for p in _list_paths(source,c, path + add) :
                    yield add + p
        elif nodetype == 'atom' :
            yield add
        else :  # Don't support negation for now
            raise Exception('Unexpected node type')



from __future__ import print_function

import math
import scipy.stats


class DistributionValue(object):

    def __init__(self):
        self.value = None

    def sample(self):
        raise NotImplementedError('Abstract method')

    def __float__(self):
        return self.sample()

    def __lt__(self, other):
        raise NotImplementedError('Abstract method')

    def __le__(self, other):
        return self.__lt__(other)

    def __eq__(self, other):
        raise NotImplementedError('Abstract method')

    def __gt__(self, other):
        return 1.0 - self.__le__(other)

    def __ge__(self, other):
        return 1.0 - self.__lt__(other)

    def constrain(self, low=None, high=None):
        """Truncate the distribution.

        :param low: lower bound (can be None)
        :param high: higher bound (can be None)
        """
        raise NotImplementedError('Abstract method')


class ScipyCDV(DistributionValue):

    def __init__(self, scipy_dist, loc=0.0, scale=1.0):
        DistributionValue.__init__(self)
        self.loc = loc
        self.scale = scale

        self.scipy_dist = scipy_dist

    def sample(self):
        return self.scipy_dist.rvs(self.loc, self.scale)

    def __lt__(self, other):
        if isinstance(other, ScipyCDV):
            other_cdf = other.scipy_dist.cdf
        else:
            other_cdf = lambda x: float(x < other)

        return self.scipy_dist.expect(other_cdf, ((), self.loc, self.scale))

    def plot(self, d=1e-5):
        import matplotlib.pyplot as plt
        import numpy

        x = numpy.linspace(self.scipy_dist.ppf(d, self.loc, self.scale),
                           self.scipy_dist.ppf(1 - d, self.loc, self.scale), 100)
        return plt.plot(x, self.scipy_dist.pdf(x, self.loc, self.scale))

    def __add__(self, other):
        if isinstance(other, DistributionValue):
            raise TypeError('sum of distributions not supported')
        else:
            return self.__class__(self.loc + other, self.scale)

    def __sub__(self, other):
        if isinstance(other, DistributionValue):
            raise TypeError('subtract of distributions not supported')
        else:
            return self.__class__(self.loc - other, self.scale)

    def __mul__(self, other):
        if isinstance(other, DistributionValue):
            raise TypeError('product of distributions not supported')
        else:
            return self.__class__(self.loc * other, self.scale * other)

    def __div__(self, other):
        if isinstance(other, DistributionValue):
            raise TypeError('product of distributions not supported')
        else:
            return self.__class__(self.loc / other, self.scale / other)


class NormalDistValue(ScipyCDV):

    def __init__(self, *args):
        ScipyCDV.__init__(self, scipy.stats.norm, *args)

    def __add__(self, other):
        if isinstance(other, NormalDistValue):
            # Sum of gaussians is gaussian.
            return self.__class__(self.loc + other.loc,
                                  math.sqrt(math.pow(self.scale, 2) + math.pow(other.scale, 2)))
        else:
            return ScipyCDV.__add__(other)

    def __sub__(self, other):
        if isinstance(other, NormalDistValue):
            # Difference of gaussians is gaussian.
            return self.__class__(self.loc - other.loc,
                                  math.sqrt(math.pow(self.scale, 2) + math.pow(other.scale, 2)))
        else:
            return ScipyCDV.__sub__(other)



"""
Part of the ProbLog distribution.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import print_function

from .logic import Constant, Term


def py2pl(d):
    """Translate a given Python datastructure into a Prolog datastructure."""

    if type(d) == list or type(d) == tuple:
        if type(d) == tuple:
            f = ','
            tail = py2pl(d[-1])
        else:
            f = '.'
            if not d:
                return Term('[]')
            tail = Term(f, py2pl(d[-1]), Term('[]'))
        for el in reversed(d[:-1]):
            tail = Term(f, py2pl(el), tail)
        return tail

    if type(d) == str:
        return Constant('"{}"'.format(d))

    if type(d) == int or type(d) == float:
        return Constant(d)

    if isinstance(d, Term):
        return d

    raise ValueError("Cannot convert from Python to Prolog: {} ({}).".format(d, type(d)))


def pl2py(d):
    """Translate a given Prolog datastructure into a Python datastructure."""

    if isinstance(d, Constant):
        if type(d.value) == str:
            return d.value.replace('"', '').replace("'", '')
        return d.value

    if isinstance(d, Term):
        if d.functor == "." and d.arity == 2:
            # list
            elements = []
            tail = d
            while isinstance(tail, Term) and tail.arity == 2 and tail.functor == '.':
                elements.append(pl2py(tail.args[0]))
                tail = tail.args[1]
            if str(tail) != '[]':
                elements.append(pl2py(tail))
            return elements
        elif d.functor == "," and d.arity == 2:
            # list
            elements = []
            tail = d
            while isinstance(tail, Term) and tail.arity == 2 and tail.functor == ',':
                elements.append(pl2py(tail.args[0]))
                tail = tail.args[1]
            if str(tail) != '[]':
                elements.append(pl2py(tail))
            return elements
        else:
            return d

    raise ValueError("Cannot convert from Prolog to Python: {} ({}).".format(d, type(d)))




"""
Maximum A-Posteriori inference for ProbLog (MAP)
"""

from __future__ import print_function


from .. import get_evaluatable
from ..program import PrologFile
from ..formula import LogicFormula
from ..util import Timer, format_dictionary
from .dtproblog import search_exhaustive, search_local
from ..constraint import TrueConstraint

import math


def main(argv):
    args = argparser().parse_args(argv)

    search = None

    # certain queries are given, they refer to facts
    #  we want to find assignments for these
    #  such that the overall probability is maximized
    # query facts => decisions with utility log p

    # Load the model
    model = PrologFile(args.filename)

    # Ground the model
    ground_program = LogicFormula.create_from(model)

    # Get the conditional probabilities of the query facts
    query_probs = get_evaluatable().create_from(ground_program).evaluate()

    # Extract queries and replace them with decisions
    decisions = []
    utilities = []
    decision_nodes = set()
    for qn, qi, ql in ground_program.labeled():
        node = ground_program.get_node(qi)
        prob = query_probs[qn]
        if type(node).__name__ != 'atom':
            raise Exception("Queries should be facts: '%s'" % qn)
        decisions.append((qi, qn))
        decision_nodes.add(qi)
        probability = prob
        utilities.append((qn, math.log(float(probability)) - math.log(1 - float(probability))))
    utilities = dict(utilities)

    # Add constraints that enforce the evidence
    for qn, qi in ground_program.evidence():
        ground_program.add_constraint(TrueConstraint(qi))
    ground_program.clear_evidence()

    # Process constraints (only on decisions, the others are encoded in the model)
    constraints = []
    for c in ground_program.constraints():
        if set(c.get_nodes()) & decision_nodes:
            constraints.append(c)

    # Compile the program (again)
    knowledge = get_evaluatable().create_from(ground_program)

    # Use dt-problog search to find the solution
    if search == 'local':
        result = search_local(knowledge, decisions, utilities, constraints)
    else:
        result = search_exhaustive(knowledge, decisions, utilities, constraints)

    # Print the result
    decisions, score, stats = result
    for x, v in decisions.items():
        if v:
            print (x)
        else:
            print (-x)


def argparser():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('filename')
    return parser
